import logging
import os
import sys


class AbstractIngestion:

    def __init__(self, handler: logging.Handler):
        self.handler = handler

    def attach(self, logger: logging.Logger):
        logger.addHandler(self.handler)

    def detach(self, logger: logging.Logger):
        logger.removeHandler(self.handler)


def get_logger(name=None, level=logging.DEBUG) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(level)
    return logger


class LevelBasedStreamHandler(logging.StreamHandler):
    def emit(self, record):
        # 根据日志级别选择输出流
        if record.levelno >= logging.WARNING:
            self.stream = sys.stderr
        else:
            self.stream = sys.stdout
        super().emit(record)


def file_handler(log_path):
    os.makedirs(os.path.dirname(log_path), exist_ok=True)
    return logging.FileHandler(log_path, mode='a', encoding='utf-8')



import subprocess
from subprocess import CompletedProcess

from davidkhala.utils.syntax.fs import rm


class Installer:
    def __init__(self, default_directory: str, source_py: str):
        self.dist = default_directory  # directory for executable binary
        self.spec = default_directory  # directory for *.spec
        self.work = default_directory  # directory for build/
        self.file = source_py
        self.name: str | None = None

    @property
    def name_options(self) -> list:
        return ["--name", self.name] if self.name else []

    def build(self) -> CompletedProcess:
        return subprocess.run([
            "pyinstaller",
            '--distpath', self.dist,
            '--specpath', self.spec,
            "--workpath", self.work,
            *self.name_options,
            "--onefile", self.file,
        ])

    def clean(self, force: bool):
        if self.dist != self.spec or force:
            rm(self.spec)
        if self.dist != self.work or force:
            rm(self.work)



from typing import Protocol, Any

import requests
from requests import Session
from requests.auth import HTTPBasicAuth

from davidkhala.utils.syntax.interface import ContextAware


def default_on_response(response: requests.Response) -> dict | None:
    """
    :param response:
    :return: the input response
    :raise HTTPError: if not response.ok
    """
    if response.ok:
        if response.text: return response.json()
        return None
    else:
        return response.raise_for_status()


class RequestProtocol(Protocol):
    def request(self, url, method: str, params=None, data=None, json=None) -> Any: ...


class Request(ContextAware, RequestProtocol):
    def __init__(self, auth: dict = None, on_response=default_on_response):
        self.options: dict = {"headers": {}}
        if auth is not None:
            bearer = auth.get("bearer")
            if bearer is not None:
                self.options["headers"]["Authorization"] = f"Bearer {bearer}"
                del auth["bearer"]
            else:
                self.options["auth"] = HTTPBasicAuth(auth["username"], auth["password"])
        self.session: Session | None = None
        self.on_response = on_response

    def open(self) -> bool:
        self.session = requests.Session()
        return True

    def close(self):
        self.session.close()
        del self.session

    def request(self, url, method: str, params=None, data=None, json=None) -> dict:
        if self.session:
            response = self.session.request(method, url, params=params, data=data, json=json, **self.options)
        else:
            response = requests.request(
                method, url, params=params, data=data, json=json, **self.options
            )
        return self.on_response(response)



import warnings
import functools

# Python 3.13+ has @warnings.deprecated
if hasattr(warnings, "deprecated"):
    deprecated = warnings.deprecated
else:
    # fallback for Python 3.12-
    def deprecated(reason: str):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                warnings.warn(
                    f"{func.__name__} is deprecated: {reason}",
                    DeprecationWarning,
                    stacklevel=2,
                )
                return func(*args, **kwargs)
            return wrapper
        return decorator



import numpy


def as_of(lat, lon):
    return numpy.array([lat, lon])


earth_radius = 6371000  # of earth, in meter


def xyz_of(lat, lon):
    lat = numpy.radians(lat)
    lon = numpy.radians(lon)
    x = earth_radius * numpy.cos(lat) * numpy.cos(lon)
    y = earth_radius * numpy.cos(lat) * numpy.sin(lon)
    z = earth_radius * numpy.sin(lat)
    return numpy.array([x, y, z])


def precision_cutoff(lat_or_lon: float):
    return format(lat_or_lon, '.6f')



import base64
import json
from pathlib import Path


def JSONReadable(data):
    return json.dumps(data, indent=4, sort_keys=True)


class Package:
    def __init__(self, name: str):
        import re
        assert re.match('^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$', name, re.IGNORECASE)


class Base64:
    @staticmethod
    def encode_file(file: Path):
        with open(file, "rb") as f:
            return Base64.encode(f.read())

    @staticmethod
    def encode(data:bytes):
        return base64.b64encode(data).decode('utf-8')


from dataclasses import dataclass
from enum import Enum


class NameEnum(Enum):
    @staticmethod
    def _generate_next_value_(name, *args):
        return name


@dataclass
class DataClass:
    ...



import contextlib
from abc import ABC, abstractmethod
from typing import Any, Protocol


class Serializable:

    def as_dict(self) -> dict:
        return self.__dict__

    def __str__(self):
        return str(vars(self))


class SupportsClose(Protocol):
    def close(self) -> None: ...


class Closeable(SupportsClose):
    @property
    def context(self):
        return contextlib.closing(self)


class Delegate:
    def __init__(self, client: Any) -> None:
        self.client = client

    def __getattr__(self, name):
        # Delegate unknown attributes/methods to the wrapped instance
        return getattr(self.client, name)


class ContextAware(ABC, SupportsClose):
    """provide Context Management Protocol"""

    def __enter__(self):
        assert self.open(), f"{self.__class__.__name__}::open() failed"
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @abstractmethod
    def open(self) -> bool: ...



import time
from typing import Callable, Any


def runtime_of(func:Callable[[], Any]) -> tuple[float, Any]:
    assert callable(func)
    start_time = time.time()
    r = func()
    return time.time() - start_time, r





import argparse

from davidkhala.poetry import reconfigure_python


def main():
    parser = argparse.ArgumentParser(description="poetry helper")
    subparsers = parser.add_subparsers(dest="command", required=True)
    reconfigure_parser = subparsers.add_parser("reconfigure", help="Rebase underneath Python's version")
    reconfigure_parser.add_argument("version", help="Semantic version", default="3.12.7")
    args = parser.parse_args()

    if args.command == 'reconfigure':
        reconfigure_python(args.version)


if __name__ == "__main__":
    main()



import os

from davidkhala.utils.syntax.env import APPDATA, python_paths, is_windows, is_linux, is_mac
from davidkhala.utils.syntax.path import join, home_resolve
from dotenv import dotenv_values, set_key


def pyvenv_cfg_path():
    remains = ['pypoetry', 'venv', 'pyvenv.cfg']
    if is_windows():
        return join(APPDATA['Roaming'], *remains)
    elif is_linux():
        return home_resolve('.config', *remains)
    elif is_mac():
        return home_resolve('Library', 'Application Support', *remains)


def reconfigure_python(sem_ver: str):
    _pyvenv_cfg_path = pyvenv_cfg_path()
    if not os.path.exists(_pyvenv_cfg_path):
        raise FileNotFoundError(f"{_pyvenv_cfg_path} not found")
    major, minor, _ = sem_ver.split('.')
    _paths = python_paths(major, minor)
    if is_windows():
        _command = f"{_paths['executable']} -m venv --clear --without-scm-ignore-files {join(APPDATA['Roaming'], 'pypoetry', 'venv')}"
    elif is_mac():
        _command = f"{_paths['executable']} -m venv --copies --clear --without-scm-ignore-files {home_resolve('Library', 'Application Support', 'pypoetry', 'venv')}"

    target = {
        **_paths,
        "version": sem_ver,
        "command": _command,
    }

    dotenv_values(_pyvenv_cfg_path)  # format validate purpose

    for key, value in target.items():
        set_key(_pyvenv_cfg_path, key, value, quote_mode="never")



import requests
from typing import Generator, Callable
from json import loads as json_loads

from requests import Session

from davidkhala.utils.http_request import Request as SessionRequest, RequestProtocol


def sse_default_filter(line: bytes) -> bool:
    return line and line != b'event: ping'


def as_sse(
        response: requests.Response,
        _filter: Callable[[bytes], bool] = sse_default_filter
) -> Generator[dict, None, None]:
    for line in response.iter_lines():
        if _filter(line):
            yield json_loads(line[5:].decode())


class Request(RequestProtocol):
    def __init__(self, borrow: SessionRequest):
        self.options: dict = borrow.options
        self.session: Session = borrow.session

    def request(self, url, method: str, params=None, data=None, json=None) -> requests.Response:
        return self.session.request(method, url, stream=True, params=params, data=data, json=json, **self.options)



from davidkhala.utils.syntax.format import JSONReadable
from davidkhala.utils.syntax.interface import Serializable


def read(path: str) -> str:
    with open(path, encoding="utf-8") as file:
        return file.read()


def write(path: str, data):
    with open(path, mode='w') as file:
        return file.write(data)


def write_json(data, name=None):
    if not name:
        name = str(data)
    if isinstance(data, Serializable):
        data = data.as_dict()
    write(f"{name}.json", JSONReadable(data))


def append(path: str, data):
    with open(path, mode='a') as file:
        return file.write(data)


from shutil import rmtree


def rm(path):
    rmtree(path, ignore_errors=True)



import os
import platform
from getpass import getuser

from packaging import version

from davidkhala.utils.syntax.path import join

APPDATA = {
    'Roaming': os.environ.get('APPDATA'),
    'Local': os.environ.get('LOCALAPPDATA')
}
USER = getuser()


def python_paths(major: str, minor: str) -> dict | None:
    if is_windows():
        home = join(APPDATA['Local'], 'Programs', 'Python', f"Python{major}{minor}")

        return {
            'home': home,
            'executable': join(home, 'python.exe'),
        }
    elif is_linux():
        home = '/usr/bin'
        return {
            'home': home,
            'executable': join(home, f"python{major}.{minor}"),
        }
    elif is_mac():
        home = f"/Library/Frameworks/Python.framework/Versions/{major}.{minor}/bin"
        return {
            "home": home,
            "executable": join(home, f"python{major}.{minor}"),
        }


class Version:
    """
    Python version
    """

    def __init__(self):
        self.major, self.minor, self.patch = platform.python_version_tuple()

    @staticmethod
    def sem_ver() -> str:
        return platform.python_version()

    @property
    def micro(self) -> str:
        return self.patch

    @staticmethod
    def is_older_than(target_version):
        """
        Check if the current Python version is older than the specified target version.

        :param target_version: A string representing the target version (e.g., "3.8.0").
        """
        current_version = Version.sem_ver()

        return version.parse(current_version) < version.parse(target_version)


def is_windows() -> bool:
    return os.name == 'nt' or platform.system() == 'Windows'


def is_linux() -> bool:
    return platform.system() == 'Linux'


def is_mac() -> bool:
    return platform.system() == 'Darwin'



from functools import reduce
from typing import Callable, TypeVar, Generic

ValueType = TypeVar('ValueType')


class Array(list, Generic[ValueType]):
    def map(self, func: Callable):
        return Array(map(func, self))

    def reduce(self, func: Callable, initial=None):
        if initial is None:
            return reduce(func, self)
        return reduce(func, self, initial)

    def filter(self, func: Callable):
        return Array(filter(func, self))

    def forEach(self, func: Callable[[ValueType, int, "Array"], None]):
        for index, value in enumerate(self):
            func(value, index, self)

    def minus(self, array):
        return [x for x in self if x not in array]



from sqlparse import split, format


class SQL:
    def __init__(self, sql: str):
        self._sql = sql

    def split(self) -> list[str]:
        return split(format(self._sql, strip_comments=True))



import os
from pathlib import Path


def join(*path_tokens):
    return os.path.join(path_tokens[0], *path_tokens[1:])


def resolve(__file__, relative_path: str = ""):
    return (Path(__file__) / relative_path).resolve()


def homedir():
    return os.path.expanduser('~')


HOME = homedir()


def dirname(file):
    return os.path.dirname(file)


def home_resolve(*path_tokens):
    return join(HOME, *path_tokens)


def delete(_path):
    Path(_path).unlink(True)



def build_url(domain, port=None, username=None, password=None, protocol="https") -> str:
    endpoint = f"{domain}:{port}" if port else domain
    auth = f"{username}:{password}@" if password else ""
    return f"{protocol}://{auth}{endpoint}"


def build(http, https) -> dict:
    return {k: v for k, v in [('http', http), ('https', https)] if v is not None}
